/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is
 * Netscape Communications Corporation.
 * Portions created by the Initial Developer are Copyright (C) 1998
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include <sys/types.h>
#include <sys/wait.h>

#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include <string>
#include <signal.h>
#include <X11/Xlib.h>
#include <X11/cursorfont.h>

#include "nsCOMPtr.h"

// for plugins
#ifndef XP_UNIX
#include "nsString.h"
#include "nsIDOMNavigator.h"
#include "nsIDOMPluginArray.h"
#include "nsIDOMPlugin.h"
#include <plugin/nsIPluginHost.h>
#include "nsIDOMMimeType.h"
#include "nsIObserverService.h"

static NS_DEFINE_CID(kPluginManagerCID, NS_PLUGINMANAGER_CID);
#endif

#define DWORD uint32_t
#define BOOL bool
#include "controller.h"
#include <fstream>
#include "debug.h"

#include <log4cpp/PropertyConfigurator.hh>
#include <log4cpp/BasicConfigurator.hh>

const char* const LOGGER_CONFIG = "/usr/share/spice/logger.ini";
char szTruststore[26];
Dl_info diDLLInfo;
int g_LookForMe[0];
std::string conf_path;

#include "plugin.h"
#include "nsScriptablePeer.h"
#include "nsIServiceManager.h"
#include "nsISupportsUtils.h" // some usefule macros are defined here

#define DWORD uint32_t
#define BOOL bool
#include "controller.h"

#define MIME_TYPES_HANDLED  "application/x-spice"
#define PLUGIN_NAME         "Spice Firefox Plugin"
#define MIME_TYPES_DESCRIPTION  MIME_TYPES_HANDLED":qsc:"PLUGIN_NAME
#define PLUGIN_DESCRIPTION  PLUGIN_NAME " Spice Client wrapper for firefox"

char* NPP_GetMIMEDescription(void)
{
    return(MIME_TYPES_DESCRIPTION);
}

//////////////////////////////////////
//
// general initialization and shutdown
//
NPError NS_PluginInitialize()
{
  return NPERR_NO_ERROR;
}

void NS_PluginShutdown()
{
}

// get values per plugin
NPError NS_PluginGetValue(NPPVariable aVariable, void *aValue)
{
  NPError err = NPERR_NO_ERROR;
  switch (aVariable) {
    case NPPVpluginNameString:
      *((char **)aValue) = PLUGIN_NAME;
      break;
    case NPPVpluginDescriptionString:
      *((char **)aValue) = PLUGIN_DESCRIPTION;
      break;
    default:
      err = NPERR_INVALID_PARAM;
      break;
  }
  return err;
}


/////////////////////////////////////////////////////////////
//
// construction and destruction of our plugin instance object
//
nsPluginInstanceBase * NS_NewPluginInstance(nsPluginCreateData * aCreateDataStruct)
{
    if(!aCreateDataStruct)
        return NULL;

    nsPluginInstance * plugin = new nsPluginInstance(aCreateDataStruct->instance);

    // now is the time to tell Mozilla that we are windowless
    NPN_SetValue(aCreateDataStruct->instance, NPPVpluginWindowBool, NULL);

    return plugin;
}

void NS_DestroyPluginInstance(nsPluginInstanceBase * aPlugin)
{
  if(aPlugin)
    delete (nsPluginInstance *)aPlugin;
}

namespace {
    char *stringCopy(const std::string &src)
    {
      char* dest = (char *) NPN_MemAlloc(src.length() + 1);
      if (dest) {
        strcpy(dest, src.c_str());
      }
      return dest;
    }
}

////////////////////////////////////////
//
// nsPluginInstance class implementation
//
nsPluginInstance::nsPluginInstance(NPP aInstance) : nsPluginInstanceBase(),
  mInstance(aInstance),
  mInitialized(false),
  mScriptablePeer(NULL)
{
    strncpy(mHomeDir, getenv("HOME"), sizeof(mHomeDir));
    strncat(mHomeDir, "/.spice", sizeof(mHomeDir));
    mkdir(mHomeDir, S_IRWXU);

    mString[0] = '\0';
    LOGGER_SECTION("spice.plugin");

    conf_path = LOGGER_CONFIG;

    std::ifstream log_init(conf_path.c_str());
    if (log_init.good()) {
        log_init.close();
        log4cpp::PropertyConfigurator::configure(conf_path.c_str());
    } else {
        log4cpp::BasicConfigurator::configure();
    }

    LOG_DEBUG("logger init");

    m_connected_status = -2;
}

nsPluginInstance::~nsPluginInstance()
{
  // mScriptablePeer may be also held by the browser 
  // so releasing it here does not guarantee that it is over
  // we should take precaution in case it will be called later
  // and zero its mPlugin member
  if (mScriptablePeer)
  	NPN_ReleaseObject(mScriptablePeer);
}

NPBool nsPluginInstance::init(NPWindow* aWindow)
{
    mInitialized = true;

    m_host_ip.clear();
    m_port.clear();
    m_password.clear();
    m_secure_port.clear();
    m_cipher_suite.clear();
    m_ssl_channels.clear();
    m_trust_store.clear();
    m_host_subject.clear();
    m_title.clear();
    m_dynamic_menu.clear();
    m_number_of_monitors.clear();
    m_guest_host_name.clear();

    m_FullScreen = PR_FALSE;
    m_AdminConsole = PR_FALSE;
    m_NoTaskMgrExecution = PR_FALSE;
    m_SendCtrlAltdelete = PR_FALSE;
    m_UsbPort = 0;
    m_UsbAutoShare = PR_FALSE;

	mWindow = aWindow; 

    return mInitialized;
}

NPError nsPluginInstance::SetWindow(NPWindow* aWindow)
{
    // keep window parameters
    mWindow = aWindow;

    return NPERR_NO_ERROR;
}


void nsPluginInstance::shut()
{
  mInitialized = false;
}

NPBool nsPluginInstance::isInitialized()
{
  return mInitialized;
}

/* attribute string strHost; */
char *nsPluginInstance::GetHostIP() const {
    return stringCopy(m_host_ip);
}
void nsPluginInstance::SetHostIP(const char * aHostIP){
    m_host_ip = aHostIP;
}

/* attribute string strPort; */
char *nsPluginInstance::GetPort() const {
    return stringCopy(m_port);
}
void nsPluginInstance::SetPort(const char * aPort) {
    m_port = aPort;
}

/* attribute ing SecurePort; */
char *nsPluginInstance::GetSecurePort() const {
    return stringCopy(m_secure_port);
}
void nsPluginInstance::SetSecurePort(const char * aSecurePort) {
    m_secure_port = aSecurePort;
}


/* attribute string strPassword; */
char *nsPluginInstance::GetPassword() const {
    return stringCopy(m_password);
}
void nsPluginInstance::SetPassword(const char * aPassword) {
    m_password = aPassword;
}

/* attribute ing Port; */
char *nsPluginInstance::GetCipherSuite() const {
    return stringCopy(m_cipher_suite);
}
void nsPluginInstance::SetCipherSuite(const char * aCipherSuite) {
    m_cipher_suite = aCipherSuite;
}

/* attribute string strPort; */
char *nsPluginInstance::GetSSLChannels() const {
    return stringCopy(m_ssl_channels);
}
void nsPluginInstance::SetSSLChannels(const char * aSSLChannels) {
    m_ssl_channels = aSSLChannels;
}

//* attribute string TrustStore; */
char *nsPluginInstance::GetTrustStore() const {
    return stringCopy(m_trust_store);
}
void nsPluginInstance::SetTrustStore(const char * aTrustStore) {
    m_trust_store = aTrustStore;
}

//* attribute string HostSubject; */
char *nsPluginInstance::GetHostSubject() const {
    return stringCopy(m_host_subject);
}
void nsPluginInstance::SetHostSubject(const char * aHostSubject) {
    m_host_subject = aHostSubject;
}

/* attribute string strFullScreen; */
PRBool nsPluginInstance::GetFullScreen() const {
    return m_FullScreen;
}
void nsPluginInstance::SetFullScreen(PRBool aFullScreen) {
    m_FullScreen = aFullScreen;
}

/* attribute string strTitle; */
char *nsPluginInstance::GetTitle() const {
    return stringCopy(m_title);
}
void nsPluginInstance::SetTitle(const char * aTitle) {
    m_title = aTitle;
}

/* attribute string strDynamicMenu; */
char *nsPluginInstance::GetDynamicMenu() const {
    return stringCopy(m_dynamic_menu);
}
void nsPluginInstance::SetDynamicMenu(const char * aDynamicMenu) {
    m_dynamic_menu = aDynamicMenu;
}

/* attribute string strNumberOfMonitors; */
char *nsPluginInstance::GetNumberOfMonitors() const {
    return stringCopy(m_number_of_monitors);
}
void nsPluginInstance::SetNumberOfMonitors(const char * aNumberOfMonitors) {
    m_number_of_monitors = aNumberOfMonitors;
}

/* attribute string strAdminConsole; */
PRBool nsPluginInstance::GetAdminConsole() const {
    return m_AdminConsole;
}
void nsPluginInstance::SetAdminConsole(PRBool aAdminConsole) {
    m_AdminConsole = aAdminConsole;
}

/* attribute string strGuestHostName; */
char *nsPluginInstance::GetGuestHostName() const {
    return stringCopy(m_guest_host_name);
}
void nsPluginInstance::SetGuestHostName(const char * aGuestHostName) {
    m_guest_host_name = aGuestHostName;
}

/* attribute string strHotKeys; */
char *nsPluginInstance::GetHotKeys() const {
    return stringCopy(m_hot_keys);
}
void nsPluginInstance::SetHotKeys(const char * aHotKeys) {
    m_hot_keys = aHotKeys;
}

/* attribute string strNoTaskMgrExecution; */
PRBool nsPluginInstance::GetNoTaskMgrExecution() const {
    return m_NoTaskMgrExecution;
}
void nsPluginInstance::SetNoTaskMgrExecution(PRBool aNoTaskMgrExecution) {
    m_NoTaskMgrExecution = aNoTaskMgrExecution;
}

/* attribute string strSendCtrlAltdelete; */
PRBool nsPluginInstance::GetSendCtrlAltdelete() const {
    return m_SendCtrlAltdelete;
}
void nsPluginInstance::SetSendCtrlAltdelete(PRBool aSendCtrlAltdelete) {
    m_SendCtrlAltdelete = aSendCtrlAltdelete;
}

/* attribute string UsbListenPort; */
unsigned short nsPluginInstance::GetUsbListenPort() const {
    return m_UsbPort;
}

void nsPluginInstance::SetUsbListenPort(unsigned short aUsbPort) {
    m_UsbPort = aUsbPort;
}

/* attribute boolean UsbAutoShare; */
PRBool nsPluginInstance::GetUsbAutoShare() const {
    return m_UsbAutoShare;
}

void nsPluginInstance::SetUsbAutoShare(PRBool aUsbAutoShare) {
    m_UsbAutoShare = aUsbAutoShare;
}

void *WaitThread (void *opaque) {
    int nRC;
    nsresult rv;
    nsPluginInstance *plugin = (nsPluginInstance *)opaque;

    waitpid(plugin->m_child_pid, &nRC, 0);
    LOG_DEBUG("spicec exit code = " << nRC);
    plugin->m_connected_status = plugin->m_external_controller->TranslateRC(nRC);
	unlink (szTruststore);
    return NULL;
}

void nsPluginInstance::Connect() {
	LOG_DEBUG ("entry");
    nsresult rv;
    PRUint32 aLength;
#ifndef XP_UNIX
    nsString path;
    nsString name;
#endif
    bool nFound = false;
	MessageData mdData;

#ifdef XP_UNIX
    // Place socket file to ~/.spice/spice-XXXXXX
    strncpy(mdData.name, mHomeDir, sizeof(mHomeDir));
    strcat(mdData.name, "/spice-XXXXXX");
    mktemp(mdData.name);
    if (setenv("SPICE_XPI_SOCKET", mdData.name, 1)) {
        LOG_ERROR ("could not set SPICE_XPI_SOCKET env variable");
        return;
    }          
               
        m_child_pid = fork();
        LOG_DEBUG("m_child_pid = " << m_child_pid);
        if (m_child_pid == 0) {
            execl (
                "/usr/libexec/spicec",
                "/usr/libexec/spicec",
                m_host_ip.c_str(), 
                m_port.c_str(), 
                m_secure_port.c_str(),
                "--controller",
                "--logger-ini",
                conf_path.c_str(),
                NULL
            );
            // failed to connect
            exit(0);
        } else {
            m_external_controller = new SpiceController(mdData);

            if (m_external_controller->Connect(10) != 0) {
                LOG_ERROR ("could not connect to spice client controller");
                return;
            }

            ControlMessage msg;
            DWORD nSize;
            int nSpicecRC;
            pthread_t wait_thread_id;

            // Send host
            msg.id = MSG_host;
            sprintf (msg.szData, "%s", m_host_ip.c_str());
            m_external_controller->Write (&msg, sizeof (msg), &nSize);

            // Send port
            msg.id = MSG_port;
            sprintf (msg.szData, "%s", m_port.c_str());
            m_external_controller->Write (&msg, sizeof (msg), &nSize);

            // Send sport
            msg.id = MSG_sport;
            sprintf (msg.szData, "%s", m_secure_port.c_str());
            m_external_controller->Write (&msg, sizeof (msg), &nSize);

            // Send FullScreen
            msg.id = MSG_full_screen;
            msg.szData[0] = m_FullScreen;
            m_external_controller->Write (&msg, sizeof (msg), &nSize);

            // Send Password
            msg.id = MSG_password;
            sprintf (msg.szData, "%s", m_password.c_str());
            m_external_controller->Write (&msg, sizeof (msg), &nSize);

            // Send cipher
            msg.id = MSG_cipher;
            sprintf (msg.szData, "%s", m_cipher_suite.c_str());
            m_external_controller->Write (&msg, sizeof (msg), &nSize);

            // Send title
            msg.id = MSG_title;
            sprintf (msg.szData, "%s", m_title.c_str());
            m_external_controller->Write (&msg, sizeof (msg), &nSize);

            // Send channels
            msg.id = MSG_encrypt_channels;
            sprintf (msg.szData, "%s", m_ssl_channels.c_str());
            m_external_controller->Write (&msg, sizeof (msg), &nSize);

            // send TruststoreFilename
            FILE *fp;
			int fd = -1;
			strncpy (
				szTruststore,  
				"/tmp/truststore.pem-XXXXXX", 
				sizeof(szTruststore)
			);
			fd = mkstemp(szTruststore);
			if (fd != -1) {
            	fp = fdopen (fd,"w+");
				if (fp != NULL) {
	            	fputs(m_trust_store.c_str(), fp);
					fclose(fp);
	            	close (fd);
				} else {
	                LOG_ERROR ("could not open truststore temp file");
    	            return;
				}
			} else {
                LOG_ERROR ("could not create truststore temp file. errno=" << errno);
                return;
			}

            msg.id = MSG_CA_file;
            sprintf (msg.szData, "%s", szTruststore);
            m_external_controller->Write (&msg, sizeof(msg), &nSize);

            // send HostSubject
            msg.id = MSG_host_cert_subject;
            memcpy (msg.szData, m_host_subject.c_str(), sizeof (msg.szData));
            m_external_controller->Write (&msg, sizeof (msg), &nSize);

            // Send HotKeys
            msg.id = MSG_hotkeys;
            sprintf (msg.szData, "%s", m_hot_keys.c_str());
            m_external_controller->Write (&msg, sizeof (msg), &nSize);

            // Send auto_display_res (AdminConsole)
            msg.id = MSG_auto_display_res;
            msg.szData[0] = m_AdminConsole;
            m_external_controller->Write (&msg, sizeof (msg), &nSize);

            // Send Connect
            msg.id = MSG_CONNECT;
            m_external_controller->Write (&msg, sizeof (msg), &nSize);

            // Set connected status
            m_connected_status = -1;

            // Setup wait thread
            pthread_create (&wait_thread_id, NULL, WaitThread, (void*)this);

            ExecuteUsbCtrl();
        }
#else
    pluginHost->GetPluginCount(&aLength);

    nsIDOMPlugin **aItems = nsnull;
    aItems = new nsIDOMPlugin*[aLength];
    if (!aItems) {
        //ERROR
    }

    rv = pluginHost->GetPlugins(aLength, aItems);
    if (NS_FAILED(rv)) {
        delete [] aItems;
        // ERROR
    }
 
    nsString string;
    for (int plugin_index = 0; plugin_index < aLength && !nFound; plugin_index++)
    {
        rv = aItems[plugin_index]->GetName(name);
        if (
            !NS_FAILED(rv) &&
            NS_ConvertUTF8toUTF16(PLUGIN_NAME).Equals(name)
        ) {
            rv = aItems[plugin_index]->GetFilename(path);
            nFound = true;
        }
    }

    if (!NS_FAILED(rv) && nFound) {
        path.Truncate(path.Length() - 37);
        path.Append(NS_LITERAL_STRING("/"));
        path.Append(NS_LITERAL_STRING("r"));
        path.Append(NS_LITERAL_STRING("e"));
        path.Append(NS_LITERAL_STRING("d"));
        path.Append(NS_LITERAL_STRING("c"));

        if (fork() == 0) {
            setenv ("LD_LIBRARY_PATH", NS_ConvertUTF16toUTF8(path).get(), 1);
        path.Append(NS_LITERAL_STRING("/"));
        path.Append(NS_LITERAL_STRING("r"));
        path.Append(NS_LITERAL_STRING("e"));
        path.Append(NS_LITERAL_STRING("d"));
        path.Append(NS_LITERAL_STRING("c"));
            execl (
                NS_ConvertUTF16toUTF8(path).get(), 
                NS_ConvertUTF16toUTF8(path).get(), 
                m_strHost, 
                m_strPort, 
                strlen(m_strPassword) > 0 ? "-password" : "",
                strlen(m_strPassword) > 0 ? m_strPassword : "",
                !m_fAudio ? "-noaudio" : "",
                NULL
            );

        }
    }
#endif
}

void nsPluginInstance::show() {
    ControlMessage msg;
    DWORD nSize;

    LOG_DEBUG("sending show message");
    // Send SHOW
    msg.id = MSG_SHOW;
    sprintf (msg.szData, "show");
    m_external_controller->Write (&msg, sizeof (msg), &nSize);
}

void nsPluginInstance::disconnect() {
    if (m_child_pid != 0) {
        kill (m_child_pid, SIGTERM);
    }
}

void nsPluginInstance::ConnectedStatus(PRInt32 *_retval) {
    *_retval = m_connected_status;
}

void nsPluginInstance::SetLanguageStrings(const char * aSection, const char * aLanguage) {
    if ((aSection != NULL) && (aLanguage != NULL)) {
        if ((strlen(aSection) > 0) && (strlen(aLanguage) > 0)) {
            m_Language[aSection] = aLanguage;
        }   
    }
}

void nsPluginInstance::SetUsbFilter(const char * aUsbFilter) {
    if (aUsbFilter != NULL) {
        m_UsbFilter = aUsbFilter;
    }
}

void nsPluginInstance::ExecuteUsbCtrl()
{
    if ((m_child_pid == 0) || (m_UsbPort == 0)) {
        return;
    }

    LOG_INFO("Executing usbrdrctrl application.");

    m_UsbCtrlPid = fork();

    if (m_UsbCtrlPid == 0) {
    
        char logname[64];
        snprintf(logname, sizeof(logname), "/tmp/usbrdrctrl-%d.log", getpid());
        
        int logfd = open(logname, O_RDWR|O_CREAT|S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
        if (logfd > 0) {
            dup2(logfd, 1);
        }
        
        char conn_name[64];
        snprintf(conn_name, sizeof(conn_name), "/tmp/SpiceForeignMenu-%d.uds", m_child_pid);
        
        LOG_DEBUG("Connection name: " << conn_name);

        char port[6];
        snprintf(port, sizeof(port), "%d", m_UsbPort);

        sleep(2);
        
        // TBD: Find executable path.
        execl("/usr/bin/usbrdrctrl",
            "/usr/bin/usbrdrctrl",
            m_guest_host_name.c_str(),
            port,
            "-c",
            conn_name,
            !m_Language["USB"].empty() ? "-l" : "",
            !m_Language["USB"].empty() ? m_Language["USB"].c_str() : "",
            !m_UsbFilter.empty() ? "-f" : "",
            !m_UsbFilter.empty() ? m_UsbFilter.c_str() : "",
            (m_UsbAutoShare == PR_TRUE) ? "-a" : "",
            NULL);

        LOG_ERROR("Failed to execute usbrdrctrl: " << errno << ": " << strerror(errno));
        exit(0);
            
    } else if (m_UsbCtrlPid > 0) {
        pthread_t tid;
        pthread_create (&tid, NULL, UsbCtrlWatchDog, this);
    }
}

void* nsPluginInstance::UsbCtrlWatchDog(void *opaque)
{
    nsPluginInstance *pThis = reinterpret_cast<nsPluginInstance*>(opaque);
    
    if (pThis != NULL) {
        int exitcode;
        
        waitpid(pThis->m_UsbCtrlPid, &exitcode, 0);
        LOG_DEBUG("usbctrl was terminated with exit code " << exitcode);
        
        pThis->m_UsbCtrlPid = 0;
        
        if (exitcode != 0) {
            pThis->ExecuteUsbCtrl();
        }
    }
    
    return NULL;
}

// ==============================
// ! Scriptability related code !
// ==============================
//
// here the plugin is asked by Mozilla to tell if it is scriptable
// we should return a valid interface id and a pointer to 
// nsScriptablePeer interface which we should have implemented
// and which should be defined in the corressponding *.xpt file
// in the bin/components folder
NPError nsPluginInstance::GetValue(NPPVariable aVariable, void *aValue)
{
  NPError rv = NPERR_NO_ERROR;

  // Here we indicate that the plugin is scriptable. See this page for details:
  // https://developer.mozilla.org/en/Gecko_Plugin_API_Reference/Scripting_plugins
  if (aVariable == NPPVpluginScriptableNPObject) {
      // addref happens in getter, so we don't addref here
    *(NPObject **)aValue = getScriptablePeer();
  }
  return rv;
}

// ==============================
// ! Scriptability related code !
// ==============================
//
// this method will return the scriptable object (and create it if necessary)
NPObject* nsPluginInstance::getScriptablePeer()
{
  NPIdentifiersInit();

  if (!mScriptablePeer) {
    mScriptablePeer = NPN_CreateObject(mInstance, GET_NPOBJECT_CLASS(ScriptablePluginObject)); 
  }
  
  if (mScriptablePeer) {
    NPN_RetainObject(mScriptablePeer);
  } 
    
  return mScriptablePeer;
}
